#include "custom_adxl355_driver_spi.h"
#include <zephyr/types.h>
#include <zephyr/sys/printk.h>
#include <ncs_version.h>
#if NCS_VERSION_NUMBER >= 0x20600
#include <zephyr/internal/syscall_handler.h>
#else
#include <zephyr/syscall_handler.h>
#endif

#include <zephyr/drivers/spi.h>

#define ADXL355_SPI_READ          0x01
#define ADXL355_SPI_WRITE         0x00
#define ADXL355_ADDR(x)			  ((x) & 0xFF)
#define GET_ADXL355_TRANSF_LEN(x) (((x) >>  8) & 0x0000FF)
#define SET_ADXL355_TRANSF_LEN(x) (((x) <<  8) & 0x00FF00)
#define GET_ADXL355_RESET_VAL(x)  (((x) >> 16) & 0x0000FF)
#define SET_ADXL355_RESET_VAL(x)  (((x) << 16) & 0xFF0000)

/* ADXL355 Register Map */
#define ADXL355_DEVID_AD     (ADXL355_ADDR(0x00) | SET_ADXL355_TRANSF_LEN(1) | SET_ADXL355_RESET_VAL(0xAD))
#define ADXL355_DEVID_MST 	 (ADXL355_ADDR(0x01) | SET_ADXL355_TRANSF_LEN(1) | SET_ADXL355_RESET_VAL(0x1D))
#define ADXL355_PARTID       (ADXL355_ADDR(0x02) | SET_ADXL355_TRANSF_LEN(1) | SET_ADXL355_RESET_VAL(0xED))
#define ADXL359_PARTID       (ADXL355_ADDR(0x02) | SET_ADXL355_TRANSF_LEN(1) | SET_ADXL355_RESET_VAL(0xE9))
#define ADXL355_REVID 		 (ADXL355_ADDR(0x03) | SET_ADXL355_TRANSF_LEN(1) | SET_ADXL355_RESET_VAL(0x01))
#define ADXL355_STATUS       (ADXL355_ADDR(0x04) | SET_ADXL355_TRANSF_LEN(1) | SET_ADXL355_RESET_VAL(0x00))
#define ADXL355_FIFO_ENTRIES (ADXL355_ADDR(0x05) | SET_ADXL355_TRANSF_LEN(1) | SET_ADXL355_RESET_VAL(0x00))
#define ADXL355_TEMP         (ADXL355_ADDR(0x06) | SET_ADXL355_TRANSF_LEN(2))
#define ADXL355_XDATA        (ADXL355_ADDR(0x08) | SET_ADXL355_TRANSF_LEN(3))
#define ADXL355_YDATA        (ADXL355_ADDR(0x0B) | SET_ADXL355_TRANSF_LEN(3))
#define ADXL355_ZDATA        (ADXL355_ADDR(0x0E) | SET_ADXL355_TRANSF_LEN(3))
#define ADXL355_FIFO_DATA    (ADXL355_ADDR(0x11) | SET_ADXL355_TRANSF_LEN(1) | SET_ADXL355_RESET_VAL(0x00))
#define ADXL355_OFFSET_X     (ADXL355_ADDR(0x1E) | SET_ADXL355_TRANSF_LEN(2))
#define ADXL355_OFFSET_Y     (ADXL355_ADDR(0x20) | SET_ADXL355_TRANSF_LEN(2))
#define ADXL355_OFFSET_Z     (ADXL355_ADDR(0x22) | SET_ADXL355_TRANSF_LEN(2))
#define ADXL355_ACT_EN       (ADXL355_ADDR(0x24) | SET_ADXL355_TRANSF_LEN(1) | SET_ADXL355_RESET_VAL(0x00))
#define ADXL355_ACT_THRESH   (ADXL355_ADDR(0x25) | SET_ADXL355_TRANSF_LEN(2))
#define ADXL355_ACT_CNT      (ADXL355_ADDR(0x27) | SET_ADXL355_TRANSF_LEN(1) | SET_ADXL355_RESET_VAL(0x01))
#define ADXL355_FILTER       (ADXL355_ADDR(0x28) | SET_ADXL355_TRANSF_LEN(1) | SET_ADXL355_RESET_VAL(0x00))
#define ADXL355_FIFO_SAMPLES (ADXL355_ADDR(0x29) | SET_ADXL355_TRANSF_LEN(1) | SET_ADXL355_RESET_VAL(0x60))
#define ADXL355_INT_MAP      (ADXL355_ADDR(0x2A) | SET_ADXL355_TRANSF_LEN(1) | SET_ADXL355_RESET_VAL(0x00))
#define ADXL355_SYNC         (ADXL355_ADDR(0x2B) | SET_ADXL355_TRANSF_LEN(1) | SET_ADXL355_RESET_VAL(0x00))
#define ADXL355_RANGE        (ADXL355_ADDR(0x2C) | SET_ADXL355_TRANSF_LEN(1) | SET_ADXL355_RESET_VAL(0x81))
#define ADXL355_POWER_CTL    (ADXL355_ADDR(0x2D) | SET_ADXL355_TRANSF_LEN(1) | SET_ADXL355_RESET_VAL(0x01))
#define ADXL355_SELF_TEST    (ADXL355_ADDR(0x2E) | SET_ADXL355_TRANSF_LEN(1) | SET_ADXL355_RESET_VAL(0x00))
#define ADXL355_RESET        (ADXL355_ADDR(0x2F) | SET_ADXL355_TRANSF_LEN(1) | SET_ADXL355_RESET_VAL(0x00))

#define ADXL355_SHADOW_REGISTER_BASE_ADDR (ADXL355_ADDR(0x50) | SET_ADXL355_TRANSF_LEN(5))
#define ADXL355_MAX_FIFO_SAMPLES_VAL  0x60
#define ADXL355_SELF_TEST_TRIGGER_VAL 0x03
#define ADXL355_RESET_CODE            0x52

/*
 * At +/- 2g with 20-bit resolution, scale is given in datasheet as
 * 3.9ug/LSB = 0.0000039 * 9.80665 = 0.00003824593 m/s^2.
 * For +/- 4g range a multiplier with value 2 is used.
 * For +/-8g range, a multiplier with value 4 is used.
 */
#define ADXL355_ACC_SCALE_FACTOR_MUL  (int64_t)     38245
#define ADXL355_ACC_SCALE_FACTOR_DIV  (int32_t)1000000000

/*
 * At +/- 10g with 20-bit resolution, scale is given in datasheet as
 * 19.5ug/LSB = 0.0000195 * 9.80665 = 0.00019122967 m/s^2.
 * For +/- 4g range a multiplier with value 2 is used.
 * For +/-8g range, a multiplier with value 4 is used.
 */
#define ADXL359_ACC_SCALE_FACTOR_MUL  (int64_t)    191229
#define ADXL359_ACC_SCALE_FACTOR_DIV  (int32_t)1000000000

/*
 * The datasheet defines an intercept of 1885 LSB at 25 degC
 * and a slope of -9.05 LSB/C. The following formula can be used to find the
 * temperature:
 * Temp = ((RAW - 1885)/(-9.05)) + 25 but this doesn't follow the format of
 * the IIO which is Temp = (RAW + OFFSET) * SCALE. Hence using some rearranging
 * we get the scale as -0.110497238 and offset as -2111.25
 */
/* OFFSET = ADXL355_TEMP_OFFSET / ADXL355_TEMP_OFFSET_DIV */
#define ADXL355_TEMP_OFFSET     -211125
#define ADXL355_TEMP_OFFSET_DIV     100

/* SCALE = ADXL355_TEMP_SCALE_FACTOR / ADXL355_TEMP_SCALE_FACTOR_DIV
   We want to return the temperature in millidegrees Celsius so instead of
   0.110497238, we are going to use 110.497238 value for scale
 */
#define ADXL355_TEMP_SCALE_FACTOR     -110497238
#define ADXL355_TEMP_SCALE_FACTOR_DIV    1000000

/*
 * The datasheet defines an intercept of 1852 LSB at 25 degC
 * and a slope of -9.05 LSB/C. The following formula can be used to find the
 * temperature:
 * Temp = ((RAW - 1852)/(-9.05)) + 25 but this doesn't follow the format of
 * the IIO which is Temp = (RAW + OFFSET) * SCALE. Hence using some rearranging
 * we get the scale as -0.110497238 and offset as -2078.25
 */
/* OFFSET = ADXL359_TEMP_OFFSET / ADXL359_TEMP_OFFSET_DIV */
#define ADXL359_TEMP_OFFSET     -207825
#define ADXL359_TEMP_OFFSET_DIV     100

/* SCALE = ADXL359_TEMP_SCALE_FACTOR / ADXL359_TEMP_SCALE_FACTOR_DIV
   We want to return the temperature in millidegrees Celsius so instead of
   0.110497238, we are going to use 110.497238 value for scale
 */
#define ADXL359_TEMP_SCALE_FACTOR     -110497238
#define ADXL359_TEMP_SCALE_FACTOR_DIV    1000000

#define ADXL355_NEG_ACC_MSK        NO_OS_GENMASK(31, 20)
#define ADXL355_RANGE_FIELD_MSK    NO_OS_GENMASK( 1,  0)
#define ADXL355_ODR_LPF_FIELD_MSK  NO_OS_GENMASK( 3,  0)
#define ADXL355_HPF_FIELD_MSK      NO_OS_GENMASK( 6,  4)
#define ADXL355_INT_POL_FIELD_MSK  NO_OS_BIT(6)

#define SPIOP	SPI_WORD_SET(8) | SPI_TRANSFER_MSB

#if DT_NUM_INST_STATUS_OKAY(DT_DRV_COMPAT) == 0
#warning "Custom ADXL355 driver enabled without any devices"
#endif

struct adxl355_data {
	uint16_t x_offset;
	uint16_t y_offset;
	uint16_t z_offset;
	uint8_t fifo_samples;
	uint8_t act_cnt;
	uint16_t act_thr;
	uint8_t comm_buff[289];
} adxldata;

struct custom_adxl355_config {
	struct spi_dt_spec spi;
};


/*
 @brief The function to read-out a single ADXL355 register
 @param reg is the address of the register
 @param data is the pointer to store read data
 @param size is the size of data
 @return 0 or err code
*/
static int adxl_read_reg(const struct device *dev, uint8_t reg, uint8_t *data, int size)
{
	int err;

	uint8_t tx_buffer = reg;

	/* Set the transmit and receive buffers */
	struct spi_buf tx_spi_buf = {.buf = (void *)&tx_buffer, .len = 1};	
	struct spi_buf_set tx_spi_buf_set = {.buffers = &tx_spi_buf, .count = 1};
	struct spi_buf rx_spi_bufs = {.buf = data, .len = size};
	struct spi_buf_set rx_spi_buf_set = {.buffers = &rx_spi_bufs, .count = 1};

	const struct custom_adxl355_config *adxl355_config = dev->config;
	err = spi_transceive_dt(&adxl355_config->spi, &tx_spi_buf_set, &rx_spi_buf_set);
	if (err < 0) {
		printk("spi_transceive_dt() failed, err: %d", err);
		return err;
	}

	return 0;
}


// @brief The function to write a value to the ADXL355 register
/// @param reg is the address of the register
/// @param values is the data to write
/// @return 0 or err code
static int adxl_write_reg(const struct device *dev, uint8_t reg, uint8_t value)
{
	int err;

	const struct custom_adxl355_config *adxl355_config = dev->config;

	/* Bit7 is 0 for the write command */
	uint8_t tx_buf[] = {(reg & 0x7F), value};

	struct spi_buf tx_spi_buf = {.buf = tx_buf, .len = sizeof(tx_buf)};
	struct spi_buf_set tx_spi_buf_set = {.buffers = &tx_spi_buf, .count = 1};
	err = spi_write_dt(&adxl355_config->spi, &tx_spi_buf_set);
	if (err < 0) {
		printk("spi_write_dt() failed, err %d", err);
		return err;
	}

	return 0;
}


static int init(const struct device *dev)
{

	int err;

	const struct custom_adxl355_config *adxl355_config = dev->config;
	err = spi_is_ready_dt(&adxl355_config->spi);
	if (!err) {
		printk("Error: SPI device is not ready, err: %d", err);
		return 0;
	}
	
	/*bme_calibrationdata(dev);
	bme_write_reg(dev, CTRLHUM, 0x04);
	bme_write_reg(dev, CTRLMEAS, 0x93);	*/

	return 0;
}

int adxl_read_values(const struct device *dev)
{
	int err;

	const struct custom_adxl355_config *adxl355_config = dev->config;
	
	int32_t datax = 0, datay = 0, dataz = 0;
	float xAxis = 0.0, yAxis = 0.0, zAxis = 0.0;

	/* Store register addresses to do burst read */
	uint8_t regs[] = {ADXL355_XDATA, ADXL355_YDATA, ADXL355_ZDATA,};	//0xFF is dummy reg
	uint8_t readbuf[sizeof(regs)];

	/* Set the transmit and receive buffers */
	struct spi_buf 		tx_spi_buf 			= {.buf = (void *)&regs, .len = sizeof(regs)};
	struct spi_buf_set 	tx_spi_buf_set		= {.buffers = &tx_spi_buf, .count = 1};
	struct spi_buf 		rx_spi_bufs			= {.buf = readbuf, .len = sizeof(regs)};
	struct spi_buf_set 	rx_spi_buf_set	= {.buffers = &rx_spi_bufs, .count = 1};
	
	/* Use spi_transceive() to transmit and receive at the same time */
	err = spi_transceive_dt(&adxl355_config->spi, &tx_spi_buf_set, &rx_spi_buf_set);
	if (err < 0) {
		printk("spi_transceive_dt() failed, err: %d", err);
		return err;
	}

	/* Put the data read from registers into actual order (see datasheet) */
	datax = (readbuf[1] << 12 | 0x0F);
	datay = (readbuf[2] << 12 | 0x0F);
	dataz = (readbuf[3] << 8);
	
	/* Print the uncompensated and compensated values */
	printk("\txAxis: = %d \n",  datax);
	printk("\tyAxis   : = %d \n", datay);	
	printk("\tzAxis   : = %d \n", dataz);

	return 0;
}


static void adxl355_print(const struct device *dev)
{
	printk("Print all characteristics of ADXL355 sensor\n");

	adxl_read_values(dev);
}

static int adxl355_open(const struct device *dev)
{
	printk("Open device for ADXL355 sensor\n");

	return 0;
}

static int adxl355_close(const struct device *dev)
{
	printk("Close device for ADXL355 sensor\n");

	return 0;
}

static int adxl355_read_reg(const struct device *dev, uint8_t reg, uint8_t *data, int size)
{
	printk("Read a given register of ADXL355 sensor\n");

	return adxl_read_reg(dev, reg, data, size);
}

static int adxl355_write_reg(const struct device *dev,  uint8_t reg, uint8_t value)
{
	printk("Write to a given register of ADXL355 sensor\n");

	return adxl_write_reg(dev, reg, value);
}

static const struct custom_adxl355_driver_api custom_adxl355_api_funcs = {
	.print      = adxl355_print,
	.open       = adxl355_open,
	.write_reg  = adxl355_write_reg,
	.read_reg   = adxl355_read_reg,
	.close      = adxl355_close,
};

/* Initializes a struct adxl355_config for an instance on a SPI bus. */
#define ADXL355_CONFIG_SPI(inst)				\
	{						\
		.spi = SPI_DT_SPEC_INST_GET(inst, SPIOP, 0),	\
	}



#define CUSTOM_ADXL355_DEFINE(inst)						\
	static struct adxl355_data adxl355_data_##inst;			\
	static const struct custom_adxl355_config custom_adxl355_config_##inst = ADXL355_CONFIG_SPI(inst);	\
	DEVICE_DT_INST_DEFINE(inst,			\
				init,							\
				NULL,							\
				&adxl355_data_##inst,	\
				&custom_adxl355_config_##inst,\
				POST_KERNEL, \
				CONFIG_DRIVER_INIT_PRIORITY, \
				&custom_adxl355_api_funcs);

/* STEP 5.2 - Create the struct device for every status "okay" node in the devicetree */
DT_INST_FOREACH_STATUS_OKAY(CUSTOM_ADXL355_DEFINE)




